# cm.py
# Functions to access the Channel Master/Rovi Guide Servers
# Copyright June 2020, Timothy Goihl

# This module lets us access the Channel Master Guide Server using an Alt name
# since we cant use the epg name, we will 'end-around' and use the Alt DNS cname
# DNS response 0x25e4 A epg.channelmastertv.com
# CNAME proxy-elb-1800886808.us-west-2.elb.amazonaws.com A 34.212.152.253 A 54.148.61.14
#
# Around Nov 23, 2020 the CNAME changed to "dvrnew-elb-579886328.us-west-2.elb.amazonaws.com"
# this new server wont respond unless the Header field 'Host' is set to "epg.channelmastertv.com"
# This can be 'spoofed' by setting the Host field in requests
# newHeaders = {'Host': CM_EPG_SERVER, 'Content-type':'text/plain','Accept':'gzip, deflate'}

import json
import requests
import logging
import datetime
import socket

CM_EPG_SERVER      = "epg.channelmastertv.com"
ALT1_CM_EPG_SERVER = "proxy-elb-1800886808.us-west-2.elb.amazonaws.com"
ALT2_CM_EPG_SERVER = "dvrnew-elb-579886328.us-west-2.elb.amazonaws.com"

# Timeouts in seconds 
#REQ_TIMEOUT = 10.001   # one timeout uses same for both
REQ_TIMEOUT = (10.0, 30.0)  # Connection Timeout=10, Read Timeout = 30

def CMInitAltServer() :
    global cmserver

    # do a CNAME Lookup in case the name of the AMZN server changes
    try :
        cname, aliaslist, ipaddrlist = socket.gethostbyname_ex(CM_EPG_SERVER)
        logging.info("CM: Defaulting CM server to: %s %s", cname, ipaddrlist)
    except Exception as e:
        logging.info("CM: Error getting default CM server %s", e)
        return None
    else :
        cmserver = cname
        return cname

# defaults to Alt server so 
cmserver = ALT2_CM_EPG_SERVER

def CMSetServerUrl(serverurl = CM_EPG_SERVER) :
    global cmserver
    cname, aliaslist, ipaddrlist = socket.gethostbyname_ex(serverurl)
    logging.info("CM: Setting CM server to: %s %s", cname, ipaddrlist)
    cmserver = serverurl

def _make_baseCMurl(apiurl) :
    global cmserver
    return 'http://' + cmserver + '/' + apiurl
    
    
# CM API Links
#GET /TVlistings/v9/listings/services/postalcode/53593/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&countrycode=US&msoid=306430&format=json
CM_SERVICE_URL = "TVlistings/v9/listings/services/postalcode/"

#GET /TVlistings/v9/listings/servicedetails/serviceid/20499/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&includechannelimages=true&
#imagehorizontalresolution=92&imageverticalresolution=36&format=json
CM_SERVICEDETAILS_URL = "TVlistings/v9/listings/servicedetails/serviceid/"

#GET /TVlistings/v9/listings/linearschedule/20499/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&
#    duration=240&inprogress=true&startdate=2020-05-02T16:00:00Z&format=json
CM_LINEARSCHED_URL = "TVlistings/v9/listings/linearschedule/"

#Here is a reference to chunked encoding, but this is for the Requests Lib, not http.server
#https://requests.readthedocs.io/en/master/user/advanced/#chunk-encoded-requests

# to re-import in interactive mode...
#https://switowski.com/blog/ipython-autoreload
#https://stackoverflow.com/questions/4385839/reimporting-a-single-function-in-python
#import sys
#from importlib import reload
#reload(sys.modules['cm'])
#reimport your function

# gloabl dict to cache the CM services dictionary so that lookups will be fast

gCmServicesResult = None

# but the caller should call CMDisposeServcies() to free the memory

def CMDisposeServcies() :  # Free the list, its big
    global gCmServicesResult
    
    gCmServicesResult = None

# Return an image object for a specific channel

def CMGetChannelImageDef(reqChannel) :
    
    global gCmServicesResult
    
    if not gCmServicesResult :
        logging.warning(" CM: CMGetChannelImageDef() Could not get CM Image" )    
        
        return None
    
    ChList = gCmServicesResult['ServiceDetailsResult']['ChannelLineup']['Channels']

    # search the CM services list for a matching channel #
    for ch in ChList :
        if ch['Channel'] == reqChannel : break
    
    # was it found?  
    if ch['Channel'] != reqChannel : return None
    
    return ch['ChannelImages'][0]

# combine 2 CM calls into one since that is how it is generally used
# Cache the Service list in a global variable because this data
# will be accessed multiple times during the PiGS service list build

def CMGetServicesFromPostalCode(postalcode) :

    global gCmServicesResult
    
    gCmServicesResult = None
    
    code, sc = CMGetServiceCode(postalcode)
    if code != 200 : return code, ""
    
    code, js = CMGetServices(sc)
    if code != 200 : return code, ""
    
    # This is one big list of channels with LOTS of information for the requested zipcode/service code
    # It has channel numbers, station ids, station icon logos, etc
    
    gCmServicesResult = js   #json.loads(js)
    
    return code


#====================================================
# The 3 base CM/Rovi API calls
#====================================================

#GET /TVlistings/v9/listings/linearschedule/20499/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&
#    duration=240&inprogress=true&startdate=2020-05-02T16:00:00Z&format=json

def CMGetLinearSchedule(servicecode, startdate, duration, inprogress) :

    #'Content-type':'text/plain','Accept':'gzip, deflate'   'Accept:':'*/*\r\n',

    newHeaders = {'Host': CM_EPG_SERVER, 'Content-type':'text/plain','Accept':'gzip, deflate'}

    myparams = {'apikey': 'v4edzwwtca4e847n72j8kesw', 'locale':'en-US',  \
                'duration': duration, 'inprogress':inprogress, \
                 'startdate':startdate, 'format':'json' }

    url = _make_baseCMurl(CM_LINEARSCHED_URL) + str(servicecode) + "/info"

    # send a GET request
    try :
        r1 = requests.get(url, headers=newHeaders, params=myparams, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0, "CM Request Timeout"
    except Exception as e :
        print('e=', e)
        return 800, 0, "CM Request Exception"
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        
        # rovi error code
        if (rj['LinearScheduleResult']['Status'] == 'NPE' ) :
            cmErrorCode = int( rj['LinearScheduleResult']['Errors'][0]['ErrorCode'] )
            logging.warning(" CM: GetLinearSchdule at %s cmErr(%d) Message: %s", startdate, cmErrorCode, 'NPE' )    
            return cmErrorCode, None;
        
        return r1.status_code, rj
    
    if r1.status_code == 500 :
        logging.warning(" CM: GetServices http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, r1.text

    # some other error
    #rj = r1.json()
    return r1.status_code, None  #rj.get('code')


#GET /TVlistings/v9/listings/services/postalcode/53593/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&countrycode=US&msoid=306430&format=json HTTP/1.1

#CANADA
#GET /TVlistings/v9/listings/services/postalcode/J7K+1N5/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-CA&countrycode=CA&msoid=306430&format=json

def CMGetServiceCode(postalcode) :

#    newHeaders = {'Content-type':'text/plain','Accept':'gzip, deflate'}
    newHeaders = {'Host': CM_EPG_SERVER, 'Content-type':'text/plain','Accept':'gzip, deflate'}
    
    if len(postalcode) == 5 :
        myparams = {'apikey': 'v4edzwwtca4e847n72j8kesw', 'locale':'en-US', 'countrycode':'US', \
                    'msoid':'306430', 'format':'json' }
        url = _make_baseCMurl(CM_SERVICE_URL) + postalcode + "/info"
    elif len(postalcode) == 6 :
        myparams = {'apikey': 'v4edzwwtca4e847n72j8kesw', 'locale':'en-CA', 'countrycode':'CA', \
                    'msoid':'306430', 'format':'json' }
        url = _make_baseCMurl(CM_SERVICE_URL) + postalcode[:3]+ '+' + postalcode[-3:] + "/info"
    
    #print('CM URL:', url)
    
    # send a GET request
    try :
        r0 = requests.get(url, headers=newHeaders, params=myparams, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0, "CM Request Timeout"
    except Exception as e :
        print('e=', e)
        return 800, 0, "CM Request Exception"
    
    # There is a bug in the CM SERVICE API results.  The docs say it should return a LIST of items service.
    # http://prod-doc.rovicorp.com/mashery/index.php/TvRest/V1/Services
    # the api returns a dict{} if there is only 1 service 53593 and a list[] if there is more than 1 53211
    # so the 'Service' key can be a list or a dict!!!
    
    # I guess we could return all service codes found, then look them both up and create
    # on big combined service code.  This would handle channels that were missing in one service
    # code.
    
    if r0.status_code == 200 :
        rj=r0.json()
        
        # rovi error code
        if (rj['ServicesResult']['Status'] == 'NPE' ) :
            cmErrorCode = rj['ServicesResult']['Errors'][0]['ErrorCode'] 
            logging.warning(" CM: GetServiceCode cmErr(%s) Message: %s", cmErrorCode, 'NPE' )    
            return cmErrorCode, None;

        if len( rj['ServicesResult']['Services'] ) == 0 :
            return r0.status_code, None;

        
        svcobj = rj['ServicesResult']['Services']['Service']
        if type(svcobj) is dict :
            servicecode = svcobj['ServiceId']
        else : # it must be a list
            servicecode = svcobj[0]['ServiceId']
        #ToDo Milwaukee has 2 service codes.  which to select???  using 0
        #DVR+ must have to make the same choice with a zi of 53211
            
        logging.info(" CM: GetServiceCode http(%d)", r0.status_code )    
        return r0.status_code, servicecode
    
    # its some error, return the code, no json
    return r0.status_code, None;

#GET /TVlistings/v9/listings/servicedetails/serviceid/20499/info?
#    apikey=v4edzwwtca4e847n72j8kesw&locale=en-US&includechannelimages=true&
#imagehorizontalresolution=92&imageverticalresolution=36&format=json HTTP/1.1
#application/x-www-form-urlencoded:
#Content-Type: application/x-www-form-urlencoded
    
def CMGetServices(servicecode) :

    #'Content-type':'text/plain',
#    newHeaders = {'Accept-Encoding':'gzip,deflate'}
    newHeaders = {'Host': CM_EPG_SERVER, 'Content-type':'text/plain','Accept':'gzip, deflate'}

    myparams = {'apikey': 'v4edzwwtca4e847n72j8kesw', 'locale':'en-US',  \
                'includechannelimages':'true', 'imagehorizontalresolution':'92', \
                 'imageverticalresolution':'36', 'format':'json' }

    url = _make_baseCMurl(CM_SERVICEDETAILS_URL) + str(servicecode) + "/info"
    
    # send a GET request
    try :
        r1 = requests.get(url, headers=newHeaders, params=myparams, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0, "CM Request Timeout"
    except Exception as e :
        print('e=', e)
        return 800, 0, "CM Request Exception"
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        
        # rovi error code
        if (rj['ServiceDetailsResult']['Status'] == 'NPE' ) :
            cmErrorCode = rj['ServiceDetailsResult']['Errors'][0]['ErrorCode'] 
            logging.warning(" CM: GetServices cmErr(%s) Message: %s", cmErrorCode, 'NPE' )    
            return cmErrorCode, None;
             
        return r1.status_code, rj
    
    if r1.status_code == 500 :
        logging.warning(" CM: GetServices http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, r1.text

    # some other error
    #rj = r1.json()
    return r1.status_code, None  #rj.get('code')
